@@ -6,9 +6,9 @@ package proxyproto
 
 import (
 	"bytes"
-	"context"
 	"crypto/tls"
 	"crypto/x509"
+	"errors"
 	"fmt"
 	"io"
 	"io/ioutil"
@@ -25,22 +25,29 @@ func TestPassthrough(t *testing.T) {
 
 	pl := &Listener{Listener: l}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
-		conn.Write([]byte("ping"))
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
 		recv := make([]byte, 4)
-		_, err = conn.Read(recv)
-		if err != nil {
-			t.Fatalf("err: %v", err)
+		if _, err = conn.Read(recv); err != nil {
+			cliResult <- err
+			return
 		}
 		if !bytes.Equal(recv, []byte("pong")) {
-			t.Fatalf("bad: %v", recv)
+			cliResult <- fmt.Errorf("bad: %v", recv)
+			return
 		}
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -61,9 +68,131 @@ func TestPassthrough(t *testing.T) {
 	if _, err := conn.Write([]byte("pong")); err != nil {
 		t.Fatalf("err: %v", err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
-func TestReadHeaderTimeout(t *testing.T) {
+// TestRequiredWithReadHeaderTimeout will iterate through 3 different timeouts to see
+// whether using a REQUIRE policy for a listener would cause an error if the timeout
+// is triggerred without a proxy protocol header being defined.
+func TestRequiredWithReadHeaderTimeout(t *testing.T) {
+	for _, duration := range []int{100, 200, 400} {
+		t.Run(fmt.Sprint(duration), func(t *testing.T) {
+			start := time.Now()
+
+			l, err := net.Listen("tcp", "127.0.0.1:0")
+
+			if err != nil {
+				t.Fatalf("err: %v", err)
+			}
+
+			pl := &Listener{
+				Listener:          l,
+				ReadHeaderTimeout: time.Millisecond * time.Duration(duration),
+				Policy: func(upstream net.Addr) (Policy, error) {
+					return REQUIRE, nil
+				},
+			}
+
+			cliResult := make(chan error)
+			go func() {
+				conn, err := net.Dial("tcp", pl.Addr().String())
+				if err != nil {
+					cliResult <- err
+					return
+				}
+				defer conn.Close()
+
+				close(cliResult)
+			}()
+
+			conn, err := pl.Accept()
+			if err != nil {
+				t.Fatalf("err: %v", err)
+			}
+			defer conn.Close()
+
+			// Read blocks forever if there is no ReadHeaderTimeout and the policy is not REQUIRE
+			recv := make([]byte, 4)
+			_, err = conn.Read(recv)
+
+			if err != nil && !errors.Is(err, ErrNoProxyProtocol) && time.Since(start)-pl.ReadHeaderTimeout > 10*time.Millisecond {
+				t.Fatal("proxy proto should not be found and time should be close to read timeout")
+			}
+			err = <-cliResult
+			if err != nil {
+				t.Fatalf("client error: %v", err)
+			}
+		})
+	}
+}
+
+// TestUseWithReadHeaderTimeout will iterate through 3 different timeouts to see
+// whether using a USE policy for a listener would not cause an error if the timeout
+// is triggerred without a proxy protocol header being defined.
+func TestUseWithReadHeaderTimeout(t *testing.T) {
+	for _, duration := range []int{100, 200, 400} {
+		t.Run(fmt.Sprint(duration), func(t *testing.T) {
+			start := time.Now()
+
+			l, err := net.Listen("tcp", "127.0.0.1:0")
+
+			if err != nil {
+				t.Fatalf("err: %v", err)
+			}
+
+			pl := &Listener{
+				Listener:          l,
+				ReadHeaderTimeout: time.Millisecond * time.Duration(duration),
+				Policy: func(upstream net.Addr) (Policy, error) {
+					return USE, nil
+				},
+			}
+
+			cliResult := make(chan error)
+			go func() {
+				conn, err := net.Dial("tcp", pl.Addr().String())
+				if err != nil {
+					cliResult <- err
+					return
+				}
+				defer conn.Close()
+
+				close(cliResult)
+			}()
+
+			conn, err := pl.Accept()
+			if err != nil {
+				t.Fatalf("err: %v", err)
+			}
+			defer conn.Close()
+
+			// 2 times the ReadHeaderTimeout because the first timeout
+			// should occur (the one set on the listener) and allow for the second to follow up
+			if err := conn.SetDeadline(time.Now().Add(pl.ReadHeaderTimeout * 2)); err != nil {
+				t.Fatalf("err: %v", err)
+			}
+
+			// Read blocks forever if there is no ReadHeaderTimeout
+			recv := make([]byte, 4)
+			_, err = conn.Read(recv)
+
+			if err != nil && !errors.Is(err, ErrNoProxyProtocol) && (time.Since(start)-(pl.ReadHeaderTimeout*2)) > 10*time.Millisecond {
+				t.Fatal("proxy proto should not be found and time should be close to read timeout")
+			}
+			err = <-cliResult
+			if err != nil {
+				t.Fatalf("client error: %v", err)
+			}
+		})
+	}
+}
+
+func TestReadHeaderTimeoutIsReset(t *testing.T) {
+	const timeout = time.Millisecond * 250
+
 	l, err := net.Listen("tcp", "127.0.0.1:0")
 	if err != nil {
 		t.Fatalf("err: %v", err)
@@ -71,20 +200,55 @@ func TestReadHeaderTimeout(t *testing.T) {
 
 	pl := &Listener{
 		Listener:          l,
-		ReadHeaderTimeout: 1 * time.Millisecond,
+		ReadHeaderTimeout: timeout,
 	}
 
-	ctx, cancel := context.WithCancel(context.Background())
-	defer cancel()
+	header := &Header{
+		Version:           2,
+		Command:           PROXY,
+		TransportProtocol: TCPv4,
+		SourceAddr: &net.TCPAddr{
+			IP:   net.ParseIP("10.1.1.1"),
+			Port: 1000,
+		},
+		DestinationAddr: &net.TCPAddr{
+			IP:   net.ParseIP("20.2.2.2"),
+			Port: 2000,
+		},
+	}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
-		<-ctx.Done()
+		// Write out the header!
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		// Sleep here longer than the configured timeout.
+		time.Sleep(timeout * 2)
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+		recv := make([]byte, 4)
+		if _, err := conn.Read(recv); err != nil {
+			cliResult <- err
+			return
+		}
+		if !bytes.Equal(recv, []byte("pong")) {
+			cliResult <- fmt.Errorf("bad: %v", recv)
+			return
+		}
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -93,10 +257,203 @@ func TestReadHeaderTimeout(t *testing.T) {
 	}
 	defer conn.Close()
 
-	// Read blocks forever if there is no ReadHeaderTimeout
-	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
+	// Set our deadlines higher than our ReadHeaderTimeout
+	if err := conn.SetReadDeadline(time.Now().Add(timeout * 3)); err != nil {
+		t.Fatalf("err: %v", err)
+	}
+	if err := conn.SetWriteDeadline(time.Now().Add(timeout * 3)); err != nil {
+		t.Fatalf("err: %v", err)
+	}
 
+	recv := make([]byte, 4)
+	if _, err = conn.Read(recv); err != nil {
+		t.Fatalf("err: %v", err)
+	}
+	if !bytes.Equal(recv, []byte("ping")) {
+		t.Fatalf("bad: %v", recv)
+	}
+
+	if _, err := conn.Write([]byte("pong")); err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	// Check the remote addr
+	addr := conn.RemoteAddr().(*net.TCPAddr)
+	if addr.IP.String() != "10.1.1.1" {
+		t.Fatalf("bad: %v", addr)
+	}
+	if addr.Port != 1000 {
+		t.Fatalf("bad: %v", addr)
+	}
+
+	h := conn.(*Conn).ProxyHeader()
+	if !h.EqualsTo(header) {
+		t.Errorf("bad: %v", h)
+	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
+}
+
+// TestReadHeaderTimeoutIsEmpty ensures the default is set if it is empty.
+// Because the default is 200ms and we wait longer than that to send a message,
+// we expect the actual address and port to be returned,
+// rather than the ProxyHeader we defined.
+func TestReadHeaderTimeoutIsEmpty(t *testing.T) {
+	DefaultReadHeaderTimeout = 200 * time.Millisecond
+
+	l, err := net.Listen("tcp", "127.0.0.1:0")
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	pl := &Listener{
+		Listener: l,
+	}
+
+	header := &Header{
+		Version:           2,
+		Command:           PROXY,
+		TransportProtocol: TCPv4,
+		SourceAddr: &net.TCPAddr{
+			IP:   net.ParseIP("10.1.1.1"),
+			Port: 1000,
+		},
+		DestinationAddr: &net.TCPAddr{
+			IP:   net.ParseIP("20.2.2.2"),
+			Port: 2000,
+		},
+	}
+
+	cliResult := make(chan error)
+	go func() {
+		conn, err := net.Dial("tcp", pl.Addr().String())
+		if err != nil {
+			cliResult <- err
+			return
+		}
+		defer conn.Close()
+
+		// Sleep here longer than the configured timeout.
+		time.Sleep(250 * time.Millisecond)
+
+		// Write out the header!
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
+	}()
+
+	conn, err := pl.Accept()
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+	defer conn.Close()
+
+	recv := make([]byte, 4)
+	if _, err = conn.Read(recv); err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	// Check the remote addr
+	addr := conn.RemoteAddr().(*net.TCPAddr)
+	if addr.IP.String() == "10.1.1.1" {
+		t.Fatalf("bad: %v", addr)
+	}
+	if addr.Port == 1000 {
+		t.Fatalf("bad: %v", addr)
+	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
+}
+
+// TestReadHeaderTimeoutIsNegative does the same as above except
+// with a negative timeout. Therefore, we expect the right ProxyHeader
+// to be returned.
+func TestReadHeaderTimeoutIsNegative(t *testing.T) {
+	l, err := net.Listen("tcp", "127.0.0.1:0")
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	pl := &Listener{
+		Listener:          l,
+		ReadHeaderTimeout: -1,
+	}
+
+	header := &Header{
+		Version:           2,
+		Command:           PROXY,
+		TransportProtocol: TCPv4,
+		SourceAddr: &net.TCPAddr{
+			IP:   net.ParseIP("10.1.1.1"),
+			Port: 1000,
+		},
+		DestinationAddr: &net.TCPAddr{
+			IP:   net.ParseIP("20.2.2.2"),
+			Port: 2000,
+		},
+	}
+
+	cliResult := make(chan error)
+	go func() {
+		conn, err := net.Dial("tcp", pl.Addr().String())
+		if err != nil {
+			cliResult <- err
+			return
+		}
+		defer conn.Close()
+
+		// Sleep here longer than the configured timeout.
+		time.Sleep(250 * time.Millisecond)
+
+		// Write out the header!
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
+	}()
+
+	conn, err := pl.Accept()
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+	defer conn.Close()
+
+	recv := make([]byte, 4)
+	if _, err = conn.Read(recv); err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	// Check the remote addr
+	addr := conn.RemoteAddr().(*net.TCPAddr)
+	if addr.IP.String() != "10.1.1.1" {
+		t.Fatalf("bad: %v", addr)
+	}
+	if addr.Port != 1000 {
+		t.Fatalf("bad: %v", addr)
+	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func TestParse_ipv4(t *testing.T) {
@@ -120,25 +477,37 @@ func TestParse_ipv4(t *testing.T) {
 			Port: 2000,
 		},
 	}
+
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
 		// Write out the header!
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
 
-		conn.Write([]byte("ping"))
 		recv := make([]byte, 4)
-		_, err = conn.Read(recv)
-		if err != nil {
-			t.Fatalf("err: %v", err)
+		if _, err = conn.Read(recv); err != nil {
+			cliResult <- err
+			return
 		}
 		if !bytes.Equal(recv, []byte("pong")) {
-			t.Fatalf("bad: %v", recv)
+			cliResult <- fmt.Errorf("bad: %v", recv)
+			return
 		}
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -148,8 +517,7 @@ func TestParse_ipv4(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != nil {
+	if _, err = conn.Read(recv); err != nil {
 		t.Fatalf("err: %v", err)
 	}
 	if !bytes.Equal(recv, []byte("ping")) {
@@ -173,6 +541,10 @@ func TestParse_ipv4(t *testing.T) {
 	if !h.EqualsTo(header) {
 		t.Errorf("bad: %v", h)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func TestParse_ipv6(t *testing.T) {
@@ -197,25 +569,36 @@ func TestParse_ipv6(t *testing.T) {
 		},
 	}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
 		// Write out the header!
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
 
-		conn.Write([]byte("ping"))
 		recv := make([]byte, 4)
-		_, err = conn.Read(recv)
-		if err != nil {
-			t.Fatalf("err: %v", err)
+		if _, err = conn.Read(recv); err != nil {
+			cliResult <- err
+			return
 		}
 		if !bytes.Equal(recv, []byte("pong")) {
-			t.Fatalf("bad: %v", recv)
+			cliResult <- fmt.Errorf("bad: %v", recv)
+			return
 		}
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -225,8 +608,7 @@ func TestParse_ipv6(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != nil {
+	if _, err = conn.Read(recv); err != nil {
 		t.Fatalf("err: %v", err)
 	}
 	if !bytes.Equal(recv, []byte("ping")) {
@@ -250,6 +632,10 @@ func TestParse_ipv6(t *testing.T) {
 	if !h.EqualsTo(header) {
 		t.Errorf("bad: %v", h)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func TestAcceptReturnsErrorWhenPolicyFuncErrors(t *testing.T) {
@@ -263,12 +649,16 @@ func TestAcceptReturnsErrorWhenPolicyFuncErrors(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -279,6 +669,10 @@ func TestAcceptReturnsErrorWhenPolicyFuncErrors(t *testing.T) {
 	if conn != nil {
 		t.Fatalf("Expected no connection, got %v", conn)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func TestReadingIsRefusedWhenProxyHeaderRequiredButMissing(t *testing.T) {
@@ -291,13 +685,21 @@ func TestReadingIsRefusedWhenProxyHeaderRequiredButMissing(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
-		conn.Write([]byte("ping"))
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -307,10 +709,13 @@ func TestReadingIsRefusedWhenProxyHeaderRequiredButMissing(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != ErrNoProxyProtocol {
+	if _, err = conn.Read(recv); err != ErrNoProxyProtocol {
 		t.Fatalf("Expected error %v, received %v", ErrNoProxyProtocol, err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func TestReadingIsRefusedWhenProxyHeaderPresentButNotAllowed(t *testing.T) {
@@ -323,10 +728,12 @@ func TestReadingIsRefusedWhenProxyHeaderPresentButNotAllowed(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 		header := &Header{
@@ -342,7 +749,12 @@ func TestReadingIsRefusedWhenProxyHeaderPresentButNotAllowed(t *testing.T) {
 				Port: 2000,
 			},
 		}
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -352,10 +764,13 @@ func TestReadingIsRefusedWhenProxyHeaderPresentButNotAllowed(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != ErrSuperfluousProxyHeader {
+	if _, err = conn.Read(recv); err != ErrSuperfluousProxyHeader {
 		t.Fatalf("Expected error %v, received %v", ErrSuperfluousProxyHeader, err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 func TestIgnorePolicyIgnoresIpFromProxyHeader(t *testing.T) {
 	l, err := net.Listen("tcp", "127.0.0.1:0")
@@ -367,10 +782,12 @@ func TestIgnorePolicyIgnoresIpFromProxyHeader(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
@@ -388,17 +805,27 @@ func TestIgnorePolicyIgnoresIpFromProxyHeader(t *testing.T) {
 				Port: 2000,
 			},
 		}
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
 
-		conn.Write([]byte("ping"))
 		recv := make([]byte, 4)
-		_, err = conn.Read(recv)
-		if err != nil {
-			t.Fatalf("err: %v", err)
+		if _, err = conn.Read(recv); err != nil {
+			cliResult <- err
+			return
 		}
 		if !bytes.Equal(recv, []byte("pong")) {
-			t.Fatalf("bad: %v", recv)
+			cliResult <- fmt.Errorf("bad: %v", recv)
+			return
 		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -408,8 +835,7 @@ func TestIgnorePolicyIgnoresIpFromProxyHeader(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != nil {
+	if _, err = conn.Read(recv); err != nil {
 		t.Fatalf("err: %v", err)
 	}
 	if !bytes.Equal(recv, []byte("ping")) {
@@ -425,6 +851,10 @@ func TestIgnorePolicyIgnoresIpFromProxyHeader(t *testing.T) {
 	if addr.IP.String() != "127.0.0.1" {
 		t.Fatalf("bad: %v", addr)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func Test_AllOptionsAreRecognized(t *testing.T) {
@@ -465,13 +895,21 @@ func TestReadingIsRefusedOnErrorWhenRemoteAddrRequestedFirst(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
-		conn.Write([]byte("ping"))
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -482,10 +920,13 @@ func TestReadingIsRefusedOnErrorWhenRemoteAddrRequestedFirst(t *testing.T) {
 
 	_ = conn.RemoteAddr()
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != ErrNoProxyProtocol {
+	if _, err = conn.Read(recv); err != ErrNoProxyProtocol {
 		t.Fatalf("Expected error %v, received %v", ErrNoProxyProtocol, err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func TestReadingIsRefusedOnErrorWhenLocalAddrRequestedFirst(t *testing.T) {
@@ -498,13 +939,21 @@ func TestReadingIsRefusedOnErrorWhenLocalAddrRequestedFirst(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
-		conn.Write([]byte("ping"))
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -515,10 +964,70 @@ func TestReadingIsRefusedOnErrorWhenLocalAddrRequestedFirst(t *testing.T) {
 
 	_ = conn.LocalAddr()
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != ErrNoProxyProtocol {
+	if _, err = conn.Read(recv); err != ErrNoProxyProtocol {
 		t.Fatalf("Expected error %v, received %v", ErrNoProxyProtocol, err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
+}
+
+func TestSkipProxyProtocolPolicy(t *testing.T) {
+	l, err := net.Listen("tcp", "127.0.0.1:0")
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+
+	policyFunc := func(upstream net.Addr) (Policy, error) { return SKIP, nil }
+
+	pl := &Listener{
+		Listener: l,
+		Policy:   policyFunc,
+	}
+
+	cliResult := make(chan error)
+	ping := []byte("ping")
+	go func() {
+		conn, err := net.Dial("tcp", pl.Addr().String())
+		if err != nil {
+			cliResult <- err
+			return
+		}
+		defer conn.Close()
+
+		if _, err := conn.Write(ping); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
+	}()
+
+	conn, err := pl.Accept()
+	if err != nil {
+		t.Fatalf("err: %v", err)
+	}
+	defer conn.Close()
+
+	_, ok := conn.(*net.TCPConn)
+	if !ok {
+		t.Fatal("err: should be a tcp connection")
+	}
+	_ = conn.LocalAddr()
+	recv := make([]byte, 4)
+	if _, err = conn.Read(recv); err != nil {
+		t.Fatalf("Unexpected read error: %v", err)
+	}
+
+	if !bytes.Equal(ping, recv) {
+		t.Fatalf("Unexpected %s data while expected %s", recv, ping)
+	}
+
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func Test_ConnectionCasts(t *testing.T) {
@@ -531,13 +1040,21 @@ func Test_ConnectionCasts(t *testing.T) {
 
 	pl := &Listener{Listener: l, Policy: policyFunc}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
-		conn.Write([]byte("ping"))
+
+		if _, err := conn.Write([]byte("ping")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -563,6 +1080,10 @@ func Test_ConnectionCasts(t *testing.T) {
 	if !ok {
 		t.Fatal("err: should be a tcp connection")
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func Test_ConnectionErrorsWhenHeaderValidationFails(t *testing.T) {
@@ -574,10 +1095,12 @@ func Test_ConnectionErrorsWhenHeaderValidationFails(t *testing.T) {
 	validationError := fmt.Errorf("failed to validate")
 	pl := &Listener{Listener: l, ValidateHeader: func(*Header) error { return validationError }}
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := net.Dial("tcp", pl.Addr().String())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
@@ -595,7 +1118,12 @@ func Test_ConnectionErrorsWhenHeaderValidationFails(t *testing.T) {
 				Port: 2000,
 			},
 		}
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := pl.Accept()
@@ -605,10 +1133,13 @@ func Test_ConnectionErrorsWhenHeaderValidationFails(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 4)
-	_, err = conn.Read(recv)
-	if err != validationError {
+	if _, err = conn.Read(recv); err != validationError {
 		t.Fatalf("expected validation error, got %v", err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 type TestTLSServer struct {
@@ -673,10 +1204,12 @@ func Test_TLSServer(t *testing.T) {
 	}
 	defer s.Close()
 
+	cliResult := make(chan error)
 	go func() {
 		conn, err := tls.Dial("tcp", s.Addr(), s.TLSClientConfig)
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
@@ -694,9 +1227,17 @@ func Test_TLSServer(t *testing.T) {
 				Port: 2000,
 			},
 		}
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
 
-		conn.Write([]byte("test"))
+		if _, err := conn.Write([]byte("test")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := s.Listener.Accept()
@@ -713,6 +1254,10 @@ func Test_TLSServer(t *testing.T) {
 	if string(recv[:n]) != "test" {
 		t.Fatalf("expected \"test\", got \"%s\" %v", recv[:n], recv[:n])
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 func Test_MisconfiguredTLSServerRespondsWithUnderlyingError(t *testing.T) {
@@ -730,6 +1275,7 @@ func Test_MisconfiguredTLSServerRespondsWithUnderlyingError(t *testing.T) {
 	}
 	defer s.Close()
 
+	cliResult := make(chan error)
 	go func() {
 		// this is not a valid TLS connection, we are
 		// connecting to the TLS endpoint via plain TCP.
@@ -745,7 +1291,8 @@ func Test_MisconfiguredTLSServerRespondsWithUnderlyingError(t *testing.T) {
 
 		conn, err := net.Dial("tcp", s.Addr())
 		if err != nil {
-			t.Fatalf("err: %v", err)
+			cliResult <- err
+			return
 		}
 		defer conn.Close()
 
@@ -763,9 +1310,17 @@ func Test_MisconfiguredTLSServerRespondsWithUnderlyingError(t *testing.T) {
 				Port: 2000,
 			},
 		}
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			cliResult <- err
+			return
+		}
 
-		conn.Write([]byte("GET /foo/bar HTTP/1.1"))
+		if _, err := conn.Write([]byte("GET /foo/bar HTTP/1.1")); err != nil {
+			cliResult <- err
+			return
+		}
+
+		close(cliResult)
 	}()
 
 	conn, err := s.Listener.Accept()
@@ -775,10 +1330,13 @@ func Test_MisconfiguredTLSServerRespondsWithUnderlyingError(t *testing.T) {
 	defer conn.Close()
 
 	recv := make([]byte, 1024)
-	_, err = conn.Read(recv)
-	if err.Error() != "tls: first record does not look like a TLS handshake" {
+	if _, err = conn.Read(recv); err.Error() != "tls: first record does not look like a TLS handshake" {
 		t.Fatalf("expected tls handshake error, got %s", err)
 	}
+	err = <-cliResult
+	if err != nil {
+		t.Fatalf("client error: %v", err)
+	}
 }
 
 type testConn struct {
@@ -808,7 +1366,9 @@ func TestCopyToWrappedConnection(t *testing.T) {
 	wrappedConn := NewConn(innerConn)
 	dummySrc := &testConn{reads: 1}
 
-	io.Copy(wrappedConn, dummySrc)
+	if _, err := io.Copy(wrappedConn, dummySrc); err != nil {
+		t.Fatalf("err: %v", err)
+	}
 	if innerConn.readFromCalledWith != dummySrc {
 		t.Error("Expected io.Copy to delegate to ReadFrom function of inner destination connection")
 	}
@@ -818,7 +1378,9 @@ func TestCopyFromWrappedConnection(t *testing.T) {
 	wrappedConn := NewConn(&testConn{reads: 1})
 	dummyDst := &testConn{}
 
-	io.Copy(dummyDst, wrappedConn)
+	if _, err := io.Copy(dummyDst, wrappedConn); err != nil {
+		t.Fatalf("err: %v", err)
+	}
 	if dummyDst.readFromCalledWith != wrappedConn.conn {
 		t.Errorf("Expected io.Copy to pass inner source connection to ReadFrom method of destination")
 	}
@@ -830,7 +1392,9 @@ func TestCopyFromWrappedConnectionToWrappedConnection(t *testing.T) {
 	innerConn2 := &testConn{}
 	wrappedConn2 := NewConn(innerConn2)
 
-	io.Copy(wrappedConn1, wrappedConn2)
+	if _, err := io.Copy(wrappedConn1, wrappedConn2); err != nil {
+		t.Fatalf("err: %v", err)
+	}
 	if innerConn1.readFromCalledWith != innerConn2 {
 		t.Errorf("Expected io.Copy to pass inner source connection to ReadFrom of inner destination connection")
 	}
@@ -850,9 +1414,10 @@ func benchmarkTCPProxy(size int, b *testing.B) {
 				break
 			}
 			_, err = io.Copy(conn, conn)
-			conn.Close()
+			// Can't defer since we keep accepting on each for iteration.
+			_ = conn.Close()
 			if err != nil {
-				b.Fatalf("Failed to read entire payload: %v", err)
+				panic(fmt.Sprintf("Failed to read entire payload: %v", err))
 			}
 		}
 	}()
@@ -872,21 +1437,21 @@ func benchmarkTCPProxy(size int, b *testing.B) {
 			}
 			bConn, err := net.Dial("tcp", backend.Addr().String())
 			if err != nil {
-				b.Fatalf("failed to dial backend: %v", err)
+				panic(fmt.Sprintf("failed to dial backend: %v", err))
 			}
 			go func() {
 				_, err = io.Copy(bConn, conn)
+				_ = bConn.(*net.TCPConn).CloseWrite()
 				if err != nil {
-					b.Fatalf("Failed to proxy incoming data to backend: %v", err)
+					panic(fmt.Sprintf("Failed to proxy incoming data to backend: %v", err))
 				}
-				bConn.(*net.TCPConn).CloseWrite()
 			}()
 			_, err = io.Copy(conn, bConn)
 			if err != nil {
-				b.Fatalf("Failed to proxy data from backend: %v", err)
+				panic(fmt.Sprintf("Failed to proxy data from backend: %v", err))
 			}
-			conn.Close()
-			bConn.Close()
+			_ = conn.Close()
+			_ = bConn.Close()
 		}
 	}()
 
@@ -914,14 +1479,16 @@ func benchmarkTCPProxy(size int, b *testing.B) {
 			b.Fatalf("err: %v", err)
 		}
 		// Write out the header!
-		header.WriteTo(conn)
+		if _, err := header.WriteTo(conn); err != nil {
+			b.Fatalf("err: %v", err)
+		}
 		//send data
 		go func() {
 			_, err = conn.Write(data)
+			_ = conn.(*net.TCPConn).CloseWrite()
 			if err != nil {
-				b.Fatalf("Failed to write data: %v", err)
+				panic(fmt.Sprintf("Failed to write data: %v", err))
 			}
-			conn.(*net.TCPConn).CloseWrite()
 
 		}()
 		//receive data
